/*
 * mbox_open_file.c
 * Open a file to use as a mbox
 * libpostal project
 * libpost_raw mbox code
 * $Id: mbox_open_file.c,v 1.13 2001/12/25 15:48:51 fullermd Exp $
 */

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <post_raw.h>
#include <post_raw_p.h>
#include <post_raw_mbox_p.h>


POSTAL_CONN *
mbox_open_file(const char *fname, int locktype, int readwrite)
{
	POSTAL_CONN *toret;
	struct stat st_ret;
	int from_str, from_str2;
	int i;

	/* Allocate memory */
	toret = malloc(PCONN_SIZE);
	toret->spec.mbox = malloc(SMBOX_SIZE);

	/* Copy the filename into the SPEC_ */
	toret->spec.mbox->filename = malloc( (strlen(fname)+1) );
	strcpy(toret->spec.mbox->filename, fname);

	/*
	 * Open the file as a raw I/O stream, lock it, and get a
	 * stdio stream, since it's more fun to manipulate.
	 * Unless we're pure read-only, open RDWR and lock exclusive.
	 */
	if( (readwrite == OPEN_READONLY) )
	{
		if((toret->spec.mbox->desc = open(fname, O_RDONLY))<0)
		{
			free(toret->spec.mbox);
			free(toret);
			return(NULL);
		}
		toret->spec.mbox->file = fdopen(toret->spec.mbox->desc, "r");
		mbox_lock_read(toret, locktype);
	}
	else
	{
		if((toret->spec.mbox->desc = open(fname, O_RDWR | O_CREAT))<0)
		{
			free(toret->spec.mbox);
			free(toret);
			return(NULL);
		}
		toret->spec.mbox->file = fdopen(toret->spec.mbox->desc, "r+");
		mbox_lock_write(toret, locktype);
	}

	/* Set the permissions in a portable way */
	fchmod(toret->spec.mbox->desc, S_IRUSR | S_IWUSR);

	/* Get the stat info */
	fstat(toret->spec.mbox->desc, &st_ret);

	/*
	 * Figure out if we already have a blank line at EOF.  
	 * I *THINK* we always should, and we'll always put one
	 * ourselves, but better safe etc.
	 * If we created a new file, we can skip all this.
	 * Skip it if we're pure read-only, too.
	 */
	if( !(readwrite == OPEN_READONLY) || (st_ret.st_size > 0) )
	{
		/* XXX Do this twice, to make sure we have \n\n at EOF */
		for(i=0;i<=1;i++)
		{
			fseek(toret->spec.mbox->file, -2, SEEK_END);
			from_str = getc(toret->spec.mbox->file);
			if( ((char)from_str) != '\n')
			{
				/*
				 * Double check, just in case we've got a stupid file.
				 * If we do, we'd need more code to deal with it in
				 * a totally 'proper' manner, unfortunately,
				 * but I'm not writing that right now.
				 */
				if( ((char)from_str) != '\r')
				{
					ungetc(from_str, toret->spec.mbox->file);
					/* Go to EOF and add a newline */
					fseek(toret->spec.mbox->file, 0, SEEK_END);
					putc('\n', toret->spec.mbox->file);
					fflush(toret->spec.mbox->file);
				}
			}
			else
			{
				/*
				 * If we DO have a \n 2 chars back, make sure
				 * we have a \n one char back, too
				 */
				from_str2 = getc(toret->spec.mbox->file);
				if( ((char)from_str2) != '\n')
				{
					/*
					 * Double check, just in case we've got a stupid file.
					 * If we do, we'd need more code to deal with it in
					 * a totally 'proper' manner, unfortunately,
					 * but I'm not writing that right now.
					 */
					if( ((char)from_str2) != '\r')
					{
						ungetc(from_str2, toret->spec.mbox->file);
						ungetc(from_str, toret->spec.mbox->file);
						/* Go to EOF and add a newline */
						fseek(toret->spec.mbox->file, 0, SEEK_END);
						putc('\n', toret->spec.mbox->file);
						fflush(toret->spec.mbox->file);
					}
				}
				else
				{
					/*
					 * In this case, and only this one, we have
					 * an affirmative exit condition right now.
					 */
					fseek(toret->spec.mbox->file, 0, SEEK_END);
					break;
				}
			}
		}
		fseek(toret->spec.mbox->file, 0, SEEK_SET);
	}


	/*
	 * If we're in READONLY mode, we're already open()'d read-only,
	 * and locked shared.
	 * If we're in WRITE mode, we're already open()'d read-write,
	 * and locked exclusive.
	 * So, only in the READ case do we need to frob, since we're
	 * open()'d read-write, and locked exclusive.  Downgrade.
	 */
	if((readwrite == OPEN_READ))
	{
		/*
		 * First, we need to do some kinda ugly hackery with
		 * open()'ing another descriptor into the file to hold onto
		 * our flock()-style locks. (see last-close() rule on flock())
		 * XXX This will require some frobbing with fcntl().  See
		 * "Braindead POSIX semantics".
		 *
		 * It's important that we hold the lock(s) here, since otherwise
		 * we open up a race condition that invalidates all the trouble
		 * we just went to to make sure it's a valid-ish (ending)
		 * mbox-format file.
		 */
		int dummydesc;
		dummydesc = open(fname, O_RDWR | O_CREAT);

		/* Close out the previous file */
		fflush(toret->spec.mbox->file);
		close(toret->spec.mbox->desc);

		/* Reopen.  No need for O_CREAT */
		toret->spec.mbox->desc = open(fname, O_RDONLY);
		toret->spec.mbox->file = fdopen(toret->spec.mbox->desc, "r");
		close(dummydesc);

		/* Move lock downscale */
		mbox_lock_read(toret, locktype);
	}

	/* Return the structure, we're ready to rock */
	return(toret);
}
